@@ -35,6 +35,7 @@
     executables that use this library to further improve performance.
 -}
 
+{-# LANGUAGE BangPatterns              #-}
 {-# LANGUAGE CPP                       #-}
 {-# LANGUAGE ExistentialQuantification #-}
 {-# LANGUAGE FlexibleContexts          #-}
@@ -104,6 +105,7 @@ module Control.Foldl (
     , map
     , hashMap
     , vector
+    , vectorM
 
     -- * Utilities
     -- $utilities
@@ -118,6 +120,8 @@ module Control.Foldl (
     , _Fold1
     , premap
     , premapM
+    , prefilter
+    , prefilterM
     , Handler
     , handles
     , foldOver
@@ -127,6 +131,7 @@ module Control.Foldl (
     , foldOverM
     , folded
     , filtered
+    , groupBy
 
     -- * Re-exports
     -- $reexports
@@ -143,7 +148,10 @@ import Control.Comonad
 import Data.Foldable (Foldable)
 import Data.Functor.Identity (Identity, runIdentity)
 import Data.Functor.Contravariant (Contravariant(..))
-import Data.Monoid
+import Data.Map.Strict (Map, alter)
+import Data.Maybe (fromMaybe)
+import Data.Monoid hiding ((<>))
+import Data.Semigroup (Semigroup(..))
 import Data.Profunctor
 import Data.Sequence ((|>))
 import Data.Vector.Generic (Vector, Mutable)
@@ -178,6 +186,8 @@ import qualified Data.HashMap.Strict         as HashMap
 import qualified Data.HashSet                as HashSet
 import qualified Data.Vector.Generic         as V
 import qualified Data.Vector.Generic.Mutable as M
+import qualified VectorBuilder.Builder
+import qualified VectorBuilder.Vector
 
 {-| Efficient representation of a left fold that preserves the fold's step
     function, initial accumulator, and extraction function
@@ -202,6 +212,10 @@ instance Profunctor Fold where
     lmap = premap
     rmap = fmap
 
+instance Choice Fold where
+  right' (Fold step begin done) = Fold (liftA2 step) (Right begin) (fmap done)
+  {-# INLINE right' #-}
+
 instance Comonad (Fold a) where
     extract (Fold _ begin done) = done begin
     {-#  INLINE extract #-}
@@ -220,11 +234,15 @@ instance Applicative (Fold a) where
         in  Fold step begin done
     {-# INLINE (<*>) #-}
 
+instance Monoid b => Semigroup (Fold a b) where
+    (<>) = liftA2 mappend
+    {-# INLINE (<>) #-}
+
 instance Monoid b => Monoid (Fold a b) where
     mempty = pure mempty
     {-# INLINE mempty #-}
 
-    mappend = liftA2 mappend
+    mappend = (<>)
     {-# INLINE mappend #-}
 
 instance Num b => Num (Fold a b) where
@@ -281,7 +299,7 @@ instance Floating b => Floating (Fold a b) where
     cos = fmap cos
     {-# INLINE cos #-}
 
-    asin = fmap sin
+    asin = fmap asin
     {-# INLINE asin #-}
 
     atan = fmap atan
@@ -355,6 +373,10 @@ instance Monad m => Profunctor (FoldM m) where
     rmap = fmap
     lmap = premapM
 
+instance (Monoid b, Monad m) => Semigroup (FoldM m a b) where
+    (<>) = liftA2 mappend
+    {-# INLINE (<>) #-}
+
 instance (Monoid b, Monad m) => Monoid (FoldM m a b) where
     mempty = pure mempty
     {-# INLINE mempty #-}
@@ -416,7 +438,7 @@ instance (Monad m, Floating b) => Floating (FoldM m a b) where
     cos = fmap cos
     {-# INLINE cos #-}
 
-    asin = fmap sin
+    asin = fmap asin
     {-# INLINE asin #-}
 
     atan = fmap atan
@@ -896,12 +918,26 @@ hashMap = Fold step begin done
     done = id
 {-# INLINABLE hashMap #-}
 
+-- | Fold all values into a vector
+vector :: Vector v a => Fold a (v a)
+vector = Fold step begin done
+  where
+    begin = VectorBuilder.Builder.empty
+
+    step x a = x <> VectorBuilder.Builder.singleton a
+
+    done = VectorBuilder.Vector.build
+{-# INLINABLE vector #-}
+
 maxChunkSize :: Int
 maxChunkSize = 8 * 1024 * 1024
 
--- | Fold all values into a vector
-vector :: (PrimMonad m, Vector v a) => FoldM m a (v a)
-vector = FoldM step begin done
+{-| Fold all values into a vector
+
+    This is more efficient than `vector` but is impure
+-}
+vectorM :: (PrimMonad m, Vector v a) => FoldM m a (v a)
+vectorM = FoldM step begin done
   where
     begin = do
         mv <- M.unsafeNew 10
@@ -916,7 +952,7 @@ vector = FoldM step begin done
     done (Pair mv idx) = do
         v <- V.freeze mv
         return (V.unsafeTake idx v)
-{-# INLINABLE vector #-}
+{-# INLINABLE vectorM #-}
 
 {- $utilities
     'purely' and 'impurely' allow you to write folds compatible with the @foldl@
@@ -1097,6 +1133,39 @@ premapM f (FoldM step begin done) = FoldM step' begin done
     step' x a = step x (f a)
 {-# INLINABLE premapM #-}
 
+{-| @(prefilter f folder)@ returns a new 'Fold' where the folder's input is used
+  only when the input satisfies a predicate f
+
+  This can also be done with 'handles' (@handles (filtered f)@) but @prefilter@
+  does not need you to depend on a lens library.
+
+> fold (prefilter p folder) list = fold folder (filter p list)
+
+>>> fold (prefilter (>5) Control.Foldl.sum) [1..10]
+40
+
+>>> fold Control.Foldl.sum (filter (>5) [1..10])
+40
+-}
+prefilter :: (a -> Bool) -> Fold a r -> Fold a r
+prefilter f (Fold step begin done) = Fold step' begin done
+  where
+    step' x a = if f a then step x a else x
+{-# INLINABLE prefilter #-}
+
+{-| @(prefilterM f folder)@ returns a new 'Fold' where the folder's input is used
+  only when the input satisfies a monadic predicate f.
+
+> foldM (prefilterM p folder) list = foldM folder (filter p list)
+-}
+prefilterM :: (Monad m) => (a -> m Bool) -> FoldM m a r -> FoldM m a r
+prefilterM f (FoldM step begin done) = FoldM step' begin done
+  where
+    step' x a = do
+      use <- f a
+      if use then step x a else return x
+{-# INLINABLE prefilterM #-}
+
 {-| A handler for the upstream input of a `Fold`
 
     Any lens, traversal, or prism will type-check as a `Handler`
@@ -1165,11 +1234,15 @@ foldOver l (Fold step begin done) =
 -}
 newtype EndoM m a = EndoM { appEndoM :: a -> m a }
 
+instance Monad m => Semigroup (EndoM m a) where
+    (EndoM f) <> (EndoM g) = EndoM (f <=< g)
+    {-# INLINE (<>) #-}
+
 instance Monad m => Monoid (EndoM m a) where
     mempty = EndoM return
     {-# INLINE mempty #-}
 
-    mappend (EndoM f) (EndoM g) = EndoM (f <=< g)
+    mappend = (<>)
     {-# INLINE mappend #-}
 
 {-| A Handler for the upstream input of `FoldM`
@@ -1246,6 +1319,17 @@ filtered p k x
     | otherwise = mempty
 {-# INLINABLE filtered #-}
 
+{-| Perform a 'Fold' while grouping the data according to a specified group
+projection function. Returns the folded result grouped as a map keyed by the
+group.
+
+-}
+groupBy :: Ord g => (a -> g) -> Fold a r -> Fold a (Map g r)
+groupBy grouper (Fold f i e) = Fold f' mempty (fmap e)
+  where
+    f' !m !a = alter (\o -> Just (f (fromMaybe i o) a)) (grouper a) m
+{-# INLINABLE groupBy #-}
+
 {- $reexports
     @Control.Monad.Primitive@ re-exports the 'PrimMonad' type class
 
